@@ -1,10 +1,14 @@
 /* pptp_ctrl.c ... handle PPTP control connection.
  *                 C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: pptp_ctrl.c,v 1.29 2004/12/13 22:06:46 quozl Exp $
+ * $Id: pptp_ctrl.c,v 1.40 2011/12/19 07:15:03 quozl Exp $
  */
 
 #include <errno.h>
+#if defined (__SVR4) && defined (__sun) /* Solaris */
+#define _XPG4_2
+#define __EXTENSIONS__
+#endif /* Solaris */
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -58,8 +62,11 @@ static struct thread_specific {
 struct PPTP_CONN {
     int inet_sock;
     /* Connection States */
-    enum { 
-        CONN_IDLE, CONN_WAIT_CTL_REPLY, CONN_WAIT_STOP_REPLY, CONN_ESTABLISHED 
+    enum {
+      CONN_IDLE,
+      CONN_WAIT_CTL_REPLY, CONN_WAIT_STOP_REPLY,
+      CONN_ESTABLISHED,
+      CONN_DEAD
     } conn_state; /* on startup: CONN_IDLE */
     /* Keep-alive states */
     enum { 
@@ -171,7 +178,7 @@ int max_echo_wait = PPTP_TIMEOUT;
 
 /* Local prototypes */
 static void pptp_reset_timer(void);
-static void pptp_handle_timer(int sig);
+static void pptp_handle_timer(void);
 /* Write/read as much as we can without blocking. */
 int pptp_write_some(PPTP_CONN * conn);
 int pptp_read_some(PPTP_CONN * conn);
@@ -190,7 +197,7 @@ int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size);
 void pptp_set_link(PPTP_CONN * conn, int peer_call_id);
 
 /*** log error information in control packets *********************************/
-static void ctrlp_error( int result, int error, int cause,
+static void ctrlp_error( int result, u_int8_t error, int cause,
         const char *result_text[], int max_result)
 {
     if( cause >= 0)
@@ -235,7 +242,7 @@ static const char *ctrl_msg_types[] = {
 #define MAX_CTRLMSG_TYPE 15
          
 /*** report a sent packet ****************************************************/
-static void ctrlp_rep( void * buffer, int size, int isbuff)
+static void ctrlp_rep( void * buffer, size_t size, int isbuff)
 {
     struct pptp_header *packet = buffer;
     unsigned int type;
@@ -257,7 +264,6 @@ static void ctrlp_rep( void * buffer, int size, int isbuff)
 /* Open new pptp_connection.  Returns NULL on failure. */
 PPTP_CONN * pptp_conn_open(int inet_sock, int isclient, pptp_conn_cb callback)
 {
-    struct sigaction sigact;
     PPTP_CONN *conn;
     /* Allocate structure */
     if ((conn = malloc(sizeof(*conn))) == NULL) return NULL;
@@ -306,11 +312,9 @@ PPTP_CONN * pptp_conn_open(int inet_sock, int isclient, pptp_conn_cb callback)
     }
     /* Set up interval/keep-alive timer */
     /*   First, register handler for SIGALRM */
-    sigact.sa_handler = pptp_handle_timer;
-    sigemptyset(&sigact.sa_mask);
-    sigact.sa_flags = SA_RESTART;
+    sigpipe_create();
+    sigpipe_assign(SIGALRM);
     global.conn = conn;
-    sigaction(SIGALRM, &sigact, &global.old_sigaction);
     /* Reset event timer */
     pptp_reset_timer();
     /* all done. */
@@ -363,7 +367,7 @@ PPTP_CALL * pptp_call_open(PPTP_CONN * conn, pptp_call_cb callback,
     }
     /* fill in the phone number if it was specified */
     if (phonenr) {
-        strncpy(packet.phone_num, phonenr, sizeof(packet.phone_num));
+        strncpy((char *)packet.phone_num, phonenr, sizeof(packet.phone_num));
         packet.phone_len = strlen(phonenr);
         if( packet.phone_len > sizeof(packet.phone_num))
             packet.phone_len = sizeof(packet.phone_num);
@@ -396,9 +400,10 @@ void pptp_call_close(PPTP_CONN * conn, PPTP_CALL * call)
     /* don't check state against WAIT_DISCONNECT... allow multiple disconnect
      * requests to be made.
      */
-    pptp_send_ctrl_packet(conn, &rqst, sizeof(rqst));
-    pptp_reset_timer();
-    call->state.pns = PNS_WAIT_DISCONNECT;
+    if (pptp_send_ctrl_packet(conn, &rqst, sizeof(rqst))) {
+        pptp_reset_timer();
+        call->state.pns = PNS_WAIT_DISCONNECT;
+    }
     /* call structure will be freed when we have confirmation of disconnect. */
 }
 
@@ -431,9 +436,10 @@ void pptp_conn_close(PPTP_CONN * conn, u_int8_t close_reason)
         pptp_call_close(conn, vector_get_Nth(conn->call, i));
     /* now close connection */
     log("Closing PPTP connection");
-    pptp_send_ctrl_packet(conn, &rqst, sizeof(rqst));
-    pptp_reset_timer(); /* wait 60 seconds for reply */
-    conn->conn_state = CONN_WAIT_STOP_REPLY;
+    if (pptp_send_ctrl_packet(conn, &rqst, sizeof(rqst))) {
+        pptp_reset_timer(); /* wait 60 seconds for reply */
+        conn->conn_state = CONN_WAIT_STOP_REPLY;
+    }
     return;
 }
 
@@ -447,10 +453,20 @@ void pptp_conn_destroy(PPTP_CONN * conn)
         pptp_call_destroy(conn, vector_get_Nth(conn->call, i));
     /* notify */
     if (conn->callback != NULL) conn->callback(conn, CONN_CLOSE_DONE);
-    sigaction(SIGALRM, &global.old_sigaction, NULL);
+    sigpipe_close();
     close(conn->inet_sock);
     /* deallocate */
     vector_destroy(conn->call);
+    conn->conn_state = CONN_DEAD;
+}
+
+int pptp_conn_is_dead(PPTP_CONN * conn)
+{
+    return conn->conn_state == CONN_DEAD;
+}
+
+void pptp_conn_free(PPTP_CONN * conn)
+{
     free(conn);
 }
 
@@ -460,6 +476,7 @@ void pptp_conn_destroy(PPTP_CONN * conn)
 void pptp_fd_set(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set,
                  int * max_fd)
 {
+    int sig_fd;
     assert(conn && conn->call);
     /* Add fd to write_set if there are outstanding writes. */
     if (conn->write_size > 0)
@@ -467,6 +484,10 @@ void pptp_fd_set(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set,
     /* Always add fd to read_set. (always want something to read) */
     FD_SET(conn->inet_sock, read_set);
     if (*max_fd < conn->inet_sock) *max_fd = conn->inet_sock;
+    /* Add signal pipe file descriptor to set */
+    sig_fd = sigpipe_fd();
+    FD_SET(sig_fd, read_set);
+    if (*max_fd < sig_fd) *max_fd = sig_fd;
 }
 
 /*** handle any pptp file descriptors set in fd_set, and clear them ***********/
@@ -474,6 +495,11 @@ int pptp_dispatch(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set)
 {
     int r = 0;
     assert(conn && conn->call);
+    /* Check for signals */
+    if (FD_ISSET(sigpipe_fd(), read_set)) {
+        if (sigpipe_read() == SIGALRM) pptp_handle_timer();
+	FD_CLR(sigpipe_fd(), read_set);
+    }
     /* Check write_set could be set. */
     if (FD_ISSET(conn->inet_sock, write_set)) {
         FD_CLR(conn->inet_sock, write_set);
@@ -510,7 +536,7 @@ int pptp_write_some(PPTP_CONN * conn) {
 	    return -1;
         }
     }
-    assert(retval <= conn->write_size);
+    assert((size_t)retval <= conn->write_size);
     conn->write_size -= retval;
     memmove(conn->write_buffer, conn->write_buffer + retval, conn->write_size);
     ctrlp_rep(conn->write_buffer, retval, 0);
@@ -713,8 +739,8 @@ int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
                     reply.version = packet->version;
                     /* protocol version not supported */
                     reply.result_code = hton8(5);
-                    pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
-                    pptp_reset_timer(); /* give sender a chance for a retry */
+                    if (pptp_send_ctrl_packet(conn, &reply, sizeof(reply)))
+                        pptp_reset_timer(); /* give sender a chance for a retry */
                 } else { /* same or greater version */
                     if (pptp_send_ctrl_packet(conn, &reply, sizeof(reply))) {
                         conn->conn_state = CONN_ESTABLISHED;
@@ -821,8 +847,8 @@ int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
                 hton8(1), hton8(PPTP_GENERAL_ERROR_NONE), 0
             };
             logecho( PPTP_ECHO_RQST);
-            pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
-            pptp_reset_timer();
+            if (pptp_send_ctrl_packet(conn, &reply, sizeof(reply)))
+                pptp_reset_timer();
             break;
         }
             /* ----------- OUTGOING CALL MESSAGES ------------ */
@@ -908,9 +934,10 @@ int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
                 vector_search(conn->call, ntoh16(packet->call_id), &call);
                 if (call->callback != NULL)
                     call->callback(conn, call, CALL_CLOSE_RQST);
-                pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
-                pptp_call_destroy(conn, call);
-                log("Call closed (RQST) (call id %d)", (int) call->call_id);
+                if (pptp_send_ctrl_packet(conn, &reply, sizeof(reply))) {
+                    pptp_call_destroy(conn, call);
+                    log("Call closed (RQST) (call id %d)", (int) call->call_id);
+                }
             }
             break;
         }
@@ -918,15 +945,25 @@ int ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
         {
             struct pptp_call_clear_ntfy *packet =
                 (struct pptp_call_clear_ntfy *)buffer;
+            int i;
+            u_int16_t our_call_id;
+            u_int16_t peer_call_id = ntoh16(packet->call_id);
             log("Call disconnect notification received (call id %d)",
-                    ntoh16(packet->call_id));
-            if (vector_contains(conn->call, ntoh16(packet->call_id))) {
-                PPTP_CALL * call;
-                ctrlp_error(packet->result_code, packet->error_code,
-                        packet->cause_code, pptp_call_disc_ntfy,
-                        MAX_CALL_DISC_NTFY);
-                vector_search(conn->call, ntoh16(packet->call_id), &call);
-                pptp_call_destroy(conn, call);
+                (int) peer_call_id);
+            /* See if we can map the peer's call id to our own */
+            for (i = 0; i < vector_size(conn->call); i++) {
+                PPTP_CALL * call = vector_get_Nth(conn->call, i);
+                if (call->peer_call_id == peer_call_id) {
+                    our_call_id = call->call_id;
+                    if (vector_contains(conn->call, our_call_id)) {
+                        ctrlp_error(packet->result_code, packet->error_code,
+                            packet->cause_code, pptp_call_disc_ntfy,
+                            MAX_CALL_DISC_NTFY);
+                        vector_search(conn->call, our_call_id, &call);
+                        pptp_call_destroy(conn, call);
+                    }
+                    break;
+                }
             }
             /* XXX we could log call stats here XXX */
             /* XXX not all servers send this XXX */
@@ -1026,28 +1063,30 @@ static void pptp_reset_timer(void)
 
 
 /*** Handle keep-alive timer **************************************************/
-static void pptp_handle_timer(int sig)
+static void pptp_handle_timer(void)
 {
     int i;
     /* "Keep Alives and Timers, 1": check connection state */
     if (global.conn->conn_state != CONN_ESTABLISHED) {
-        if (global.conn->conn_state == CONN_WAIT_STOP_REPLY) 
+        if (global.conn->conn_state == CONN_WAIT_STOP_REPLY) {
             /* hard close. */
             pptp_conn_destroy(global.conn);
-        else /* soft close */
-            pptp_conn_close(global.conn, PPTP_STOP_NONE);
+            return;
+        }
+        /* soft close */
+        pptp_conn_close(global.conn, PPTP_STOP_NONE);
     }
     /* "Keep Alives and Timers, 2": check echo status */
-    if (global.conn->ka_state == KA_OUTSTANDING) /*no response to keep-alive*/
-        pptp_conn_close(global.conn, PPTP_STOP_NONE);
-    else { /* ka_state == NONE */ /* send keep-alive */
+    if (global.conn->ka_state == KA_OUTSTANDING) {
+        /* no response to keep-alive */
+        log ("closing control connection due to missing echo reply");
+	pptp_conn_close(global.conn, PPTP_STOP_NONE);
+    } else { /* ka_state == NONE */ /* send keep-alive */
         struct pptp_echo_rqst rqst = {
             PPTP_HEADER_CTRL(PPTP_ECHO_RQST), hton32(global.conn->ka_id) };
-        pptp_send_ctrl_packet(global.conn, &rqst, sizeof(rqst));
-        global.conn->ka_state = KA_OUTSTANDING;
-        /* XXX FIXME: wake up ctrl thread -- or will the SIGALRM do that
-         * automagically? XXX
-         */
+        if (pptp_send_ctrl_packet(global.conn, &rqst, sizeof(rqst))) {
+            global.conn->ka_state = KA_OUTSTANDING;
+        }
     }
     /* check incoming/outgoing call states for !IDLE && !ESTABLISHED */
     for (i = 0; i < vector_size(global.conn->call); i++) {
